{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Guide to tackling a Dynamic Programming problem with the help of code\n",
    "\n",
    "# Step 1: generate some test cases\n",
    "import random\n",
    "def problem_instance():\n",
    "    vals = 'abcdefghijklmnopqrstuvwxyz'\n",
    "    freq = [random.random() for _ in vals]\n",
    "    tot = sum(freq)\n",
    "    freq = [f / tot for f in freq] # normalize\n",
    "    return vals, freq\n",
    "vals, freq = problem_instance()\n",
    "n = len(vals)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Balanced tree: ['n', 'g', 'd', 'b', 'a', 'c', 'f', 'e', 'k', 'i', 'h', 'j', 'm', 'l', 'u', 'r', 'p', 'o', 'q', 't', 's', 'x', 'w', 'v', 'z', 'y']\n"
     ]
    }
   ],
   "source": [
    "# A binary search tree is either:\n",
    "# - None (the empty tree)\n",
    "# - (val, left, right) a value, and left/right children\n",
    "\n",
    "def balanced_tree(arr):\n",
    "    if len(arr) == 0: return None\n",
    "    k = len(arr)//2\n",
    "    return (arr[k], balanced_tree(arr[:k]), balanced_tree(arr[k+1:]))\n",
    "\n",
    "def bst_search(q, T):\n",
    "    if T is None:\n",
    "        return False\n",
    "    (Tval, Tleft, Tright) = T\n",
    "    if q == Tval: return True\n",
    "    if q <  Tval: return bst_search(q, Tleft)\n",
    "    if q >  Tval: return bst_search(q, Tright)\n",
    "    \n",
    "def preorder(T):\n",
    "    if T is None: return []\n",
    "    Tval, Tleft, Tright = T\n",
    "    return [Tval] + preorder(Tleft) + preorder(Tright)\n",
    "\n",
    "def check_bst(T):\n",
    "    for q in vals:\n",
    "        assert(bst_search(q, T) == True)\n",
    "        \n",
    "T = balanced_tree(vals)\n",
    "print('Balanced tree:', preorder(T))\n",
    "check_bst(T)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Avg cost of balanced tree: 4.020400409220588\n"
     ]
    }
   ],
   "source": [
    "def bst_search_length(q, T):\n",
    "    if T is None:\n",
    "        return 0\n",
    "    (Tval, Tleft, Tright) = T\n",
    "    if q == Tval: return 1\n",
    "    if q <  Tval: return 1 + bst_search_length(q, Tleft)\n",
    "    if q >  Tval: return 1 + bst_search_length(q, Tright)\n",
    "\n",
    "def average_cost(T):\n",
    "    avg = 0\n",
    "    for i,q in enumerate(vals):\n",
    "        avg += freq[i] * bst_search_length(q, T)\n",
    "    return avg\n",
    "\n",
    "print('Avg cost of balanced tree:', average_cost(T))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "bestT: ['j', 'e', 'b', 'a', 'c', 'd', 'g', 'f', 'i', 'h', 'u', 'p', 'm', 'l', 'k', 'n', 'o', 'r', 'q', 's', 't', 'x', 'v', 'w', 'z', 'y']\n",
      "Avg cost of bestT: 3.704721373772423\n"
     ]
    }
   ],
   "source": [
    "# Step 2: Write the recursive (and implicitly memoized) solution\n",
    "\n",
    "# Optional: Precompute the range Freq table\n",
    "# F[lo][hi] is the cumulative frequency of vals[lo] to vals[hi] inclusive\n",
    "F = [[0] * (n+1) for _ in range(n)]\n",
    "for i in range(n):\n",
    "    F[i][i] = freq[i]\n",
    "    for j in range(i+1, n):\n",
    "        F[i][j] = F[i][j-1] + freq[j]\n",
    "assert(abs(F[0][n-1] - 1.0) < 0.0001)\n",
    "# print('freq:', freq)\n",
    "# print('F:', F)\n",
    "\n",
    "class Memoize: \n",
    "    def __init__(self, func):\n",
    "        self.tbl = {}\n",
    "        self.func = func\n",
    "    def __call__(self, *args):\n",
    "        if args in self.tbl: return self.tbl[args]\n",
    "        self.tbl[args] = self.func(*args)\n",
    "        return self.tbl[args]\n",
    "\n",
    "@Memoize\n",
    "# From lo to hi inclusive\n",
    "def optbst_recursive(lo, hi):\n",
    "    if lo > hi:\n",
    "        return 0, None # Empty tree\n",
    "    bestavg = 1000000 # that's a big tree\n",
    "    bestT = None\n",
    "    for k in range(lo, hi + 1):\n",
    "        l,ltree = optbst_recursive(lo,k-1)\n",
    "        r,rtree = optbst_recursive(k+1,hi)\n",
    "        s = F[lo][hi] + l + r\n",
    "        if s < bestavg:\n",
    "            bestavg = s\n",
    "            bestT = (vals[k], ltree, rtree)\n",
    "    return bestavg, bestT\n",
    "\n",
    "bestAvg, bestT = optbst_recursive(0, n-1)\n",
    "print('bestT:', preorder(bestT))\n",
    "print('Avg cost of bestT:', average_cost(bestT))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Step 3: Work out the (explicit) dynamic programming solution\n",
    "def optbst_dynprog():\n",
    "    OPT = [[None] * (n+1) for _ in range(n)]\n",
    "    # Compute Cost\n",
    "    def compute_cost(lo, hi):\n",
    "        bestavg = 1000000 # that's a big tree\n",
    "        for k in range(lo, hi + 1):\n",
    "            l = OPT[lo][k-1] if k > lo else 0\n",
    "            r = OPT[k+1][hi] if k < hi else 0\n",
    "            s = F[lo][hi] + l + r\n",
    "            if s < bestavg:\n",
    "                bestavg = s\n",
    "            OPT[lo][hi] = bestavg\n",
    "    for lo in range(n, -1, -1):\n",
    "        for hi in range(lo, n):\n",
    "            compute_cost(lo, hi)\n",
    "            assert OPT[lo][hi] == optbst_recursive(lo, hi)[0]\n",
    "    return OPT[0][n-1]\n",
    "optval = optbst_dynprog()\n",
    "assert(optval == bestAvg)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "['n', 'g', 'd', 'b', 'a', 'c', 'f', 'e', 'k', 'i', 'h', 'j', 'm', 'l', 'u', 'r', 'p', 'o', 'q', 't', 's', 'x', 'w', 'v', 'z', 'y']\n",
      "['j', 'e', 'b', 'a', 'c', 'd', 'g', 'f', 'i', 'h', 'u', 'p', 'm', 'l', 'k', 'n', 'o', 'r', 'q', 's', 't', 'x', 'v', 'w', 'z', 'y']\n"
     ]
    }
   ],
   "source": [
    "# Step 4: Work out a dynamic programming solution that records the solution\n",
    "def optbst_dynprog2():\n",
    "    global OPT\n",
    "    OPT = [[None] * (n+1) for _ in range(n)]\n",
    "    # Compute Cost\n",
    "    def compute_cost(lo, hi):\n",
    "        bestavg = 1000000 # that's a big tree\n",
    "        bestk = -1  # Keep track of the index k with the best score\n",
    "        for k in range(lo, hi + 1):\n",
    "            l,lk = OPT[lo][k-1] if k > lo else (0, -1)\n",
    "            r,rk = OPT[k+1][hi] if k < hi else (0, -1)\n",
    "            s = F[lo][hi] + l + r\n",
    "            if s < bestavg:\n",
    "                bestavg = s\n",
    "                bestk = k    # Store the k that gave the best score\n",
    "        OPT[lo][hi] = bestavg, bestk\n",
    "    for lo in range(n,-1,-1):\n",
    "        for hi in range(lo, n):\n",
    "            compute_cost(lo, hi)\n",
    "            assert OPT[lo][hi][0] == optbst_recursive(lo, hi)[0]\n",
    "    return OPT[0][n-1]\n",
    "\n",
    "# Check the iterative solution and tree itself\n",
    "best3, T3 = optbst_dynprog2()\n",
    "assert(best3 == bestAvg)\n",
    "print(preorder(T))\n",
    "\n",
    "# Go find the solution in O(n) time\n",
    "def preorder2(lo, hi, OPT):\n",
    "    if lo > hi:\n",
    "        return []\n",
    "    _, k = OPT[lo][hi]\n",
    "    return [vals[k]] + preorder2(lo, k-1, OPT) + preorder2(k+1, hi, OPT)\n",
    "print(preorder2(0, n-1, OPT))\n",
    "assert(preorder2(0, n-1, OPT) == preorder(bestT))"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.8.3"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
